/*
             CREATOR: DAn williams
                DATE:
         DESCRIPTION: makes state machine code
            FILENAME: combo3.c
             STARTED:
    OPERATING SYSTEM: Linux
1st VERSION FINISHED:

this code writes state machine code

the states are grouped into banks that are addressed by their inputs, 
so the offset in the state machine is local wihtin the input bank

  factor x
     state 1
     state 2
     state 3
  factor y
     state a
     state b

*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>

//#define debug(k)   printf("%s\n", (k));
#define debug(k)  /* (k) */

#define REPLYSZ       512

//global variables

typedef struct FSM_s {  

  unsigned int factors;     // factors in the state machine (inputs)
  unsigned int * states;    // number of states each factor has
  
  char **  factorNames;     // names of factors
  char ***  stateNames;     // names of states
  
  unsigned int * linkTable; // link to next state (feedback)
  
} FSM_t;           

void test();
void chomp(char * string);
void postPad(char * string, int finalWidth);

void FSMInit(FSM_t * this) ;
void FSMDestroy(FSM_t * this);
void FSMAddFactor(FSM_t * this, char * name);
void FSMDump(FSM_t * this) ;
void FSMDumpFactors(FSM_t * this);
void FSMDumpStates(FSM_t * this, unsigned int factor);
long FSMFindFactor(FSM_t * this, char * name);
void FSMAddFactorState(FSM_t * this, unsigned int thisFactor, char * name);
void FSMstateTable (FSM_t * this, int * maxNameLen) ;
void FSMDumpCode(FSM_t * this);

void makeMachine();
int  toIndex(int * Limit, int LimitCount, int * CoOrd);
void fromIndex(int * Limit, int LimitCount, int * Result, int i) ;



int main(int argc, char** argv) {
  int retCode;

  
  retCode = 0;
  printf(" This program writes state machines \n");
  printf(" Written by Dan Williams aug 21 2006 \n");
  
  makeMachine();
  //test();
  
  return(retCode);
}

//------------------------------------------------------------------------

void makeMachine() {
  FSM_t * Project;
  
  char * string;
  char reply[REPLYSZ];
  int i;

  Project = NULL;
  Project = realloc(Project, sizeof(FSM_t));
  FSMInit(Project);
  
  do { 
    printf("Enter a factor name or press enter when done\n");
    string=fgets(reply,REPLYSZ,stdin);
    chomp(string);
    if (strlen(string) != 0) {
       FSMAddFactor(Project, string);
    }
  } while (strlen(string) != 0);  
  
  for(i = 0; i < Project->factors; i++) {
    do { 
      printf("Enter a state name for factor %s or press enter for next\n",  Project->factorNames[i]);
      string=fgets(reply,REPLYSZ,stdin);
      chomp(string);
      if (strlen(string) != 0) {         
         FSMAddFactorState(Project, i, string);      
      }
    } while (strlen(string) != 0);  
   }
   
   FSMDumpCode(Project);
  
   FSMDestroy(Project); 
}



void test() {
  FSM_t * Project;

  Project = NULL;
  debug("Allocating main project...");  
  Project = realloc(Project, sizeof(FSM_t));
  
  FSMInit(Project);
  FSMAddFactor(Project, "bar");
  FSMAddFactor(Project, "that");
  FSMAddFactor(Project, "other");
    
//  printf(" 'foo' is index %d\n", FSMFindFactor(Project, "bar"));
  
  FSMAddFactorState(Project, FSMFindFactor(Project, "bar"), "factor1");
  FSMAddFactorState(Project, FSMFindFactor(Project, "bar"), "fac2");
  
  FSMAddFactorState(Project, FSMFindFactor(Project, "that"), "facA");
  FSMAddFactorState(Project, FSMFindFactor(Project, "that"), "facB");
  
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2A");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2B");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "factor2C");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2D");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2E");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2F");
  
  FSMDumpCode(Project);
  
  FSMDestroy(Project); 
}

// add padding to output, for alignment
void postPad(char * string, int finalWidth){
 int i;
  for(i = 0; i < finalWidth-strlen(string); i++) { printf(" ");}
}

void chomp(char * string) {
  char * target;
  
  target = strchr(string,'\n');
  if (target != NULL) { *target = '\0'; }
  
}

void FSMInit(FSM_t * this) {

  debug("Entering init...");
  this->factors     = 0;
  this->states      = NULL;
  this->factorNames = NULL;
  this->stateNames  = NULL;
  this->linkTable   = NULL;

}

void FSMDestroy(FSM_t * this) {
  unsigned int i, j;
 
  for (i = 0; i < this->factors; i++ ) {
    free(this->factorNames[i]);  
    for (j = 0; j < this->states[i]; j++) {
      free(this->stateNames[i][j]);
    }     
  }
  free(this->factorNames);
  free(this->states);     
  free(this->stateNames);
  
  this->factors = 0;
  
}

void FSMAddFactor(FSM_t * this, char * name) {

  this->factors++;
  this->factorNames                  = realloc(this->factorNames, (sizeof(char*)*this->factors));
  this->factorNames[this->factors-1] = strdup(name);   
  this->states                       = realloc( this->states, sizeof(unsigned int)*this->factors );
  this->states[this->factors-1]      = 0;
  this->stateNames                   = realloc(this->stateNames,  sizeof(char**)*this->factors );

}

void FSMDump(FSM_t * this) {
  unsigned int i, j;
 
  for (i = 0; i < this->factors; i++ ) {
    printf("%s\n", this->factorNames[i] );
    for (j = 0; j < this->states[i]; j++) {
      printf("  %s\n", this->stateNames[i][j]);
    }
  }

}

void FSMDumpFactors(FSM_t * this) {
  unsigned int i, j;
 
  for (i = 0; i < this->factors; i++ ) {
    printf("%s\n", this->factorNames[i] );
  }

}

void FSMDumpStates(FSM_t * this, unsigned int factor) {
  unsigned int i, j;
 
    for (j = 0; j < this->states[factor]; j++) {
      printf("  %s\n", this->stateNames[factor][j]);
    }

}

long FSMFindFactor(FSM_t * this, char * name) {
  unsigned int i;
  
  for (i = 0; i < this->factors; i++ ) {
    if (strcmp(name, this->factorNames[i]) == 0)  return i;
  }
  
  return -1;
}

void FSMAddFactorState(FSM_t * this, unsigned int thisFactor, char * name) {

   this->states[thisFactor]++;       
   this->stateNames[thisFactor]                             = realloc(this->stateNames[thisFactor], (sizeof(char*)*this->states[thisFactor]));
   this->stateNames[thisFactor][this->states[thisFactor]-1] = strdup(name);    

}

void FSMDumpCode(FSM_t * this) {

   unsigned char bits, totalBits;
   int * maxNameLen;
   int i, j;
      
   maxNameLen = malloc( sizeof(int)*this->factors );

   printf("typedef struct FSM_State_s { \n");
   totalBits = 0;
   for(i = 0; i < this->factors; i++) {
     if (this->states[i] > 1) {
       for (bits = 0; (1<<bits) < this->states[i]; bits++);
       printf("  unsigned %s : %d;\n", this->factorNames[i], bits);
       totalBits += bits;
     }
   }
   printf("} FSM_State_t;\n");
   
      printf(" \n\
typedef unsigned int  PTRType; \n\
    \n\
typedef struct stateEntry_s {  \n\
    PTRType       next;            \n\
    FSM_State_t   state;           \n\
 } stateEntry_t;                   \n\
 \n");
 
   // figure out padding for string alignment
    for(i = 0; i < this->factors; i++) {   
      maxNameLen[i] = 0;
      for (j = 0; j < this->states[i]; j++) {   
        if (this->states[i] > 1) {          
          if (strlen(this->stateNames[i][j]) > maxNameLen[i])  maxNameLen[i] = strlen(this->stateNames[i][j]);
        }
      }   
    }
 
   // make C code enumerations
   for(i = 0; i < this->factors; i++) {   
     printf ("//----------| %s |----------\n", this->factorNames[i]);
     for (j = 0; j < this->states[i]; j++) {   
       if (this->states[i] > 1) {          
         printf ("  #define %s ", this->stateNames[i][j]);
         postPad(this->stateNames[i][j], maxNameLen[i]);
         printf("%d\n", j) ;        
       }
     }   
   }
           
    FSMstateTable (this, maxNameLen);
        
}



void FSMstateTable (FSM_t * this, int * maxNameLen) {

       int i,j;
       int * stateItter;
       int entries;
       int stateMask;

       stateItter = malloc( sizeof(int)*this->factors );
       entries = 1;       
       stateMask = (1<<this->factors)-1;
       
       for (i = 0; i < this->factors; i++ )  entries *= this->states[i];
       
       printf("//------------\n  //  ");
       for(i = this->factors - 1; i >= 0; i--) {
         if (this->states[i] > 1) printf ("  %s  ,", this->factorNames[i]);  
       }
       printf("\n");
              
       printf("stateEntry_t stateTable[] = {   \n");
       
       for (i = 0; i < entries; i++) {
       
         fromIndex(this->states, this->factors, stateItter, i);
                 
         printf("  { %3d, {", i & stateMask); 
         for(j = this->factors - 1; j >= 0; j--) {
           if (this->states[j] > 1) {
             printf (" %s  ", this->stateNames[j][stateItter[j]]);     
             postPad(this->stateNames[j][stateItter[j]], maxNameLen[j]);
             printf (",");  
           }           
         }       
         printf(" }},  /*   index: %u   */",i); 
           printf("\n");
       }
       printf("};");
       free(stateItter);
   
}


// forward and reverse table navigation

void fromIndex(int * Limit, int LimitCount, int * Result, int i) {
  int j, k, n;

  for(j = LimitCount - 1; j >= 0; j--) {
   for (k = 0, n = 1; k < j; k++) n *= Limit[k];
   Result[j] = i / n;
   i -= (Result[j] * n);   
   
  }
}


int toIndex(int * Limit, int LimitCount, int * CoOrd){
  int i;
  int r;
  int m;
  
  m = 1; 
  r = 0;
   
  for( i = 0; i < LimitCount; i++) {
    r += CoOrd[i] * m;
    m *= Limit[i];
  }   

 return r;
 
}
