/*
             CREATOR: DAn williams
                DATE:
         DESCRIPTION: makes state machine code
            FILENAME: combo3.c
             STARTED:
    OPERATING SYSTEM: Linux
1st VERSION FINISHED:

this code writes state machine code

the states are grouped into banks that are addressed by their inputs, 
so the offset in the state machine is local wihtin the input bank

  factor x
     state 1
     state 2
     state 3
  factor y
     state a
     state b

*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

//#define debug(k)   printf("%s\n", (k));
#define debug(k)  /* (k) */

#define REPLYSZ       512

#define INPUT    0
#define OUTPUT   1
#define FEEDBACK 2

char * DIRECTIONS[] = {"in", "out", "fb" };

#define VERSION 6.2

//global variables


typedef struct FSM_s {  

  char          * filename;   // filename

  unsigned int  factors;      // factors in the state machine (inputs / feedback)
  unsigned int  * states;     // number of states each factor has
  unsigned char * directions; // input or output
  
  char **  factorNames;       // names of factors
  char ***  stateNames;       // names of states
  
  unsigned int * linkTable;   // link to next state (feedback)
  
} FSM_t;           

/*

from factor state factor state factor state
to   factor state factor state

fromfactors
fromstates
tofactors
tostates




*/

void test();
void chomp(char * string);
void postPad(char * string, int finalWidth);

void FSMInit(FSM_t * this) ;
void FSMDestroy(FSM_t * this);
void FSMAddFactor(FSM_t * this, char * name, int direction);
void FSMDelFactorState(FSM_t * this, unsigned int factor, unsigned int state) ;
long FSMFindState(FSM_t * this, unsigned int factor, char * name) ;
void FSMDelFactor(FSM_t * this, unsigned int factor) ; 
void FSMDump(FSM_t * this) ;
void FSMDumpFactors(FSM_t * this);
void FSMDumpStates(FSM_t * this, unsigned int factor);
long FSMFindFactor(FSM_t * this, char * name);
void FSMAddFactorState(FSM_t * this, unsigned int thisFactor, char * name);
void FSMstateTable (FSM_t * this, int * maxNameLen) ;
void FSMDumpCode(FSM_t * this);

void FSMDumpRom(FSM_t * this);  // writeme
void FSMinterlinkTable(FSM_t * this);  //writeme

void makeMachine();
int  doCmd(char * string, FSM_t * Project);
int  getCommand(char ** commands, char ** in) ;

void cmdHelp();
void cmdAddFactor( FSM_t * Project, char * params);
void cmdDelFactor( FSM_t * Project, char * params);
void cmdAddState( FSM_t * Project, char * params);
void cmdDelState( FSM_t * Project, char * params);
void cmdShowStates( FSM_t * Project, char * params);
void cmdFileName( FSM_t * Project, char * params);
void cmdSave( FSM_t * Project, char * params);
void cmdLoad( FSM_t * Project, char * params);


int  toIndex(int * Limit, int LimitCount, int * CoOrd);
void fromIndex(int * Limit, int LimitCount, int * Result, int i) ;

void printHexString( char * s);



int main(int argc, char** argv) {
  int retCode;
  
  if (isatty(STDIN_FILENO)) {   
    // terminal
  } else {
    // pipe
  }
  
  retCode = 0;
  printf(" This program writes state machines \n");
  printf(" Written by Dan Williams Feb 24 2007 \n");
  printf(" Ver. %f \n", VERSION);
  
  makeMachine();
  //test();
  
  return(retCode);
}

//------------------------------------------------------------------------

// interact with stdin to make a machine
void makeMachine() {

  FSM_t * Project;  
  char * string;
  char reply[REPLYSZ];

  Project = NULL;
  Project = realloc(Project, sizeof(FSM_t));
  FSMInit(Project);
  
  printf(" Type \"help\" for help. \n");
  do {   
    printf("> ");
    fflush(stdout);
    string = fgets(reply, REPLYSZ, stdin);      
  } while((string != NULL) && (!doCmd(string, Project))); 
    
   FSMDestroy(Project); 
}



// Interpert command string
int doCmd(char * string, FSM_t * Project) {

// hey doc, these gotta match!
  enum cmdnums { 
         EMPTY = 0,
         HELP,
         ADDFACTOR,
         ADDSTATE,
         BUILD,
         QUIT,
         EXIT,
         SHOW,
         SHOWFACTORS,
         SHOWSTATES,
         FILENAME,
         SAVE,
         LOAD,
         DELFACTOR,
         DELSTATE,
         NEW,
         UNKNOWN
         
  };  
  
  static char * commands[] = { 
         " ",
         "help", 
         "addfactor", 
         "addstate", 
         "build", 
         "quit",
         "exit",
         "show",
         "showfactors",
         "showstates",
         "filename",
         "save",
         "load",
         "delfactor",
         "delstate",
         "new",
         "" 
  };
  
    chomp(string); 
       
    switch( getCommand(commands, &string)) {
      case HELP:
         cmdHelp();
         break;
      case BUILD:
         FSMDumpCode(Project);
         break;
      case ADDFACTOR:  
         cmdAddFactor(Project, string);      
         break;
      case ADDSTATE:
         cmdAddState(Project, string);          
         break;  
      case EXIT:          
      case QUIT:
         return 1;
         break;
      case SHOW:
         FSMDump(Project);  
         break;
      case SHOWFACTORS:
         FSMDumpFactors(Project);    
         break;     
      case SHOWSTATES:
         cmdShowStates(Project, string); 
         break;
      case FILENAME:
         cmdFileName(Project, string);
         break;
      case SAVE:
         cmdSave(Project, string);
         break;
      case LOAD:
         cmdLoad(Project, string);
         break;    
      case NEW:
         FSMDestroy(Project);
         FSMInit(Project);
         break;  
      case DELSTATE:
         cmdDelState(Project, string);
         break;
      case DELFACTOR:
         cmdDelFactor(Project, string);
         break;
      case UNKNOWN:
         printf("unrecognized command.\n");
         break;
    
    }
    
    return 0;

}


// determine which command was entered
int getCommand(char ** commands, char ** in) {

  char * cursor;
  char * sep = " \t";
  int  i; 

 while( (**in == ' ') || (**in == '\t')) (*in)++;  // skip white space 
 if (**in == '\0') return 0;               // 0 is EMPTY
 cursor = strsep(in, sep); 
 for( i = 0; ((strcmp(cursor, commands[i]) !=0 ) && (*commands[i] != '\0')) ; i++); 
 return i;
 
}

void printHexString( char * s) {
  for(;*s ;s++) printf("%02X ", *s);
}

// set project filename
void cmdFileName( FSM_t * Project, char * params) {
  char * cursor;   
  char * sep = " \t";

  debug("entering cmdfilename");
 
  if ((params==NULL) || (strlen(params) == 0)) {
    if (Project->filename != NULL) 
       printf("%s\n", Project->filename);
    return;
  } else {  
    if (Project->filename != params) { // we tend to get called with our own pointer
      if (Project->filename != NULL) {
        free(Project->filename);      
      } 
      Project->filename = strdup(params);
    }
  } 
  
}

// save machine to file
void cmdSave( FSM_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   FILE * output;
   unsigned int i, j;
   
  debug("entering cmdsave"); 
  if ((params==NULL) || (strlen(params) == 0) ) {
     if (Project->filename != NULL) {
       debug("calling self with filename");
       cmdSave( Project, Project->filename);
     } else {
       printf ("I need you to give me a filename!\n");
     }
     return;
  } else {   
    debug("seperating commands");
    while( cursor = strsep(&params, sep) ) { 
      if ((output = fopen(cursor, "wb")) == NULL) {
        printf("Couldn't open %s to save machine.\n", cursor);
      } else {
        for (i = 0; i < Project->factors; i++ ) {
          debug("adding factor");
          fprintf(output, "addfactor %s %s\n",  DIRECTIONS[Project->directions[i]], Project->factorNames[i]);
          for (j = 0; j < Project->states[i]; j++) {
            debug("adding state");
            fprintf(output, "  addstate %s %s\n", Project->factorNames[i], Project->stateNames[i][j]);
            debug("done adding state");
          }
        }
        debug("closing file");
        fclose(output);
        debug("resetting filename");
        cmdFileName(Project, cursor);
      }                   
    }
  }           
}


// load file into current machine
void cmdLoad( FSM_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   FILE * input;
   char * string;
   char reply[REPLYSZ];
   
  if ((params==NULL) || (strlen(params) == 0)) {
    if (Project->filename != NULL) {
       cmdLoad( Project, Project->filename);
     } else {
       printf ("I need you to give me a filename!\n");
     }
     return;
  } else {   
    while( cursor = strsep(&params, sep) ) { 
      if ((input = fopen(cursor, "rb")) == NULL) {
        printf("Couldn't open %s to load machine.\n", cursor);
      } else {      
        do {   
          string = fgets(reply, REPLYSZ, input);      
        } while((string != NULL) && (!doCmd(string, Project)));            
        fclose(input);
        cmdFileName(Project, cursor);
      }                   
    }    
  }           
}


void cmdShowStates( FSM_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   int i;

   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need to know which factor to show you states of!\n");
      return;
   } else {   
      while( cursor = strsep(&params, sep) ) { 
         if ((i = FSMFindFactor(Project, cursor)) == -1) {
           printf ("I cannot find a factor called '%s'.\n", cursor);
         } else {
           printf("%s:\n", cursor);
           FSMDumpStates(Project, i);
         }
      }
   }           
}


void cmdAddState( FSM_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   int i;

   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need to know which factor to add your state to!\n");
      return;
   } else {
      cursor = strsep(&params, sep);
      if ((i = FSMFindFactor(Project, cursor)) == -1) {
        printf ("I cannot find a factor called '%s'.\n", cursor);
        return;
      }
   }
   
   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need to know what state names you want to add to %s.\n", cursor);
      return;
   }
   
   while( cursor = strsep(&params, sep) ) {      
      FSMAddFactorState(Project, i, cursor); 
   }
}



void cmdDelState( FSM_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   int i, j;

   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need to know which factor to delete your state from!\n");
      return;
   } else {
      cursor = strsep(&params, sep);
      if ((i = FSMFindFactor(Project, cursor)) == -1) {
        printf ("I cannot find a factor called '%s'.\n", cursor);
        return;
      }
   }
   
   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need to know what state names you want to delete from %s.\n", cursor);
      return;
   }
   
   while( cursor = strsep(&params, sep) ) {     
      if ((j = FSMFindState( Project, i, cursor)  ) == -1) {
        printf ("I cannot find a state called %s in factor called '%s'.\n", cursor, Project->factorNames[i]);
        return;  
      } else {
        FSMDelFactorState(Project, i, j); 
      }
   }
}



void cmdAddFactor( FSM_t * Project, char * params) {
   char * cursor;
   char * sep = " \t";
   char dir;

   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need a direction and name for the factor you want to add!\n");
      return;
   }
   
   cursor = strsep(&params, sep);
   
   if (strcmp(cursor, "in") == 0) {
     dir = INPUT;
   } else if ( strcmp(cursor, "out") == 0) {
     dir = OUTPUT;
   } else if ( strcmp(cursor, "fb") == 0) {
     dir = FEEDBACK;
   } else {
     printf("direction needs to be 'in', 'fb' or 'out', you said %s.\n", cursor);
     return;
   }
      
   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need a name for the factor you want to add!\n");
      return;
   }   
      
   while( cursor = strsep(&params, sep) ) {
      FSMAddFactor(Project, cursor, dir);
   }
}



void cmdDelFactor( FSM_t * Project, char * params) {
   char * cursor;
   char * sep = " \t";
   char dir;
   int i;

   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need a  name for the factor you want to delete!\n");
      return;
   }  
   
   while( cursor = strsep(&params, sep) ) {   
      if ((i = FSMFindFactor( Project, cursor) ) == -1) {
        printf ("I cannot find a factor called '%s'.\n", cursor );
        return;  
      } else {
        FSMDelFactor(Project, i); 
      }      
   }
}


void cmdHelp() {

printf( "\
help                         print help (your looking at it)    \n\
addfactor direction  name(s) add a factor                       \n\
  direction is 'in', 'fb' or 'out'                              \n\
addstate  factor  name       add a state to factor              \n\
showfactors                  list factors                       \n\
showstates factor            list states                        \n\
show                         show all factors and states        \n\
build                         print machine                     \n\
quit                         exit                               \n\
filename {filename}          report or set the current filename \n\
save {filename}                save machine definition          \n\
load {filename}                load state machine definition    \n\
new                          clear project                      \n\
                                                                \n\
delfactor factor             delete factor                      \n\
delstate factor state        delete state from factor           \n\
                                                                \n\
           {} denote optional paramiters.                       \n\
");
}





// add padding to output, for alignment
void postPad(char * string, int finalWidth){
 int i;
  for(i = 0; i < finalWidth-strlen(string); i++) { printf(" ");}
}

// tear off any trailing linefeeds
void chomp(char * string) {
  char * target;
  
  target = strchr(string,'\n');
  if (target != NULL) { *target = '\0'; }
  
}


//------------------------------------------------- FSM --------------------------------------------------------


// initialize a new machine
void FSMInit(FSM_t * this) {

  debug("Entering init...");
  this->factors     = 0;
  this->states      = NULL;
  this->directions  = NULL;
  this->factorNames = NULL;
  this->stateNames  = NULL;
  this->linkTable   = NULL;  
  this->filename    = NULL;

}

// deallocate a machine
void FSMDestroy(FSM_t * this) {
  unsigned int i, j;
 
  for (i = 0; i < this->factors; i++ ) {
    free(this->factorNames[i]);      
    for (j = 0; j < this->states[i]; j++) {
      free(this->stateNames[i][j]);
    }     
  }
  free(this->factorNames); this->factorNames = NULL;
  free(this->states);      this->states      = NULL;
  free(this->stateNames);  this->stateNames  = NULL;
  free(this->directions);  this->directions  = NULL;  
  free(this->filename);    this->filename    = NULL;
  free(this->linkTable);   this->linkTable   = NULL;
  
  this->factors = 0;
  
}

/*
  add a factor to a machine
  Oh really, add some error checking already.
*/
void FSMAddFactor(FSM_t * this, char * name, int direction) {

  this->factors++;                                                                                   // another factor
  this->factorNames                  = realloc(this->factorNames, (sizeof(char*)*this->factors));    // another name
  this->factorNames[this->factors-1] = strdup(name);                                                 // set name
  this->states                       = realloc( this->states, sizeof(unsigned int)*this->factors );  // another state count 
  this->states[this->factors-1]      = 0;                                                            // with 0 states
  this->stateNames                   = realloc(this->stateNames, (sizeof(char**)*this->factors));    // another array of state names
  this->stateNames[this->factors-1]  = NULL;                                                         // with nothing in it
  this->directions                   = realloc(this->directions, (sizeof(char)*this->factors));      // another direction
  this->directions[this->factors-1]  = direction;                                                    // that is ~

}


void FSMDelFactorState(FSM_t * this, unsigned int factor, unsigned int state) {

 unsigned int i, j;
 
  if (state <= this->states[factor]) {
     for (j = state; j < this->states[factor] - 1; j++) {
          this->stateNames[factor][j] = this->stateNames[factor][j+1]; 
     }
     
     free(this->stateNames[factor][j+1]);
     this->states[factor]--; 
               
  } 
}


long FSMFindState(FSM_t * this, unsigned int factor, char * name) {
  unsigned int i, j;
  for (j = 0; j < this->states[factor]; j++) {
    if (strcmp(name, this->stateNames[factor][j]) == 0)  return j;
  }
  return -1;
}



void FSMDelFactor(FSM_t * this, unsigned int factor) {
  unsigned int i, j;
  if (factor <= this->factors) {
    free(this->factorNames[factor]);                   // delete its name
    
    for (j = 0; j < this->states[factor]; j++) {       // delete its factors
      free(this->stateNames[factor][j]);
    }
  
    for (j = factor; j < this->factors - 1; j++) {
      // move back the forward states
      this->states[j] = this->states[j+1];   
      
      // move back the forward directions
      this->directions[j] = this->directions[j+1]; 
      
      // move back the forward factorNames
      this->factorNames[j] = this->factorNames[j+1]; 
      
      // move back the forward stateNames
      this->stateNames[j] = this->stateNames[j+1];
    }
    
    this->factors--;
    
    this->factorNames                  = realloc(this->factorNames, (sizeof(char*)*this->factors)); 
    this->states                       = realloc(this->states,      (sizeof(unsigned int)*this->factors ));
    this->stateNames                   = realloc(this->stateNames,  (sizeof(char**)*this->factors));
    this->directions                   = realloc(this->directions,  (sizeof(char)*this->factors)); 
    
  }
}


// dump the states and factors for a machine
void FSMDump(FSM_t * this) {
  unsigned int i, j;
 
  for (i = 0; i < this->factors; i++ ) {
    printf("%s  (%s)\n", this->factorNames[i], DIRECTIONS[this->directions[i]] );
    for (j = 0; j < this->states[i]; j++) {
      printf("  %s\n", this->stateNames[i][j]);
    }
  }

}

// dump the factors for a machine
void FSMDumpFactors(FSM_t * this) {
  unsigned int i, j;
 
  for (i = 0; i < this->factors; i++ ) {
    printf("%s\n", this->factorNames[i] );
  }

}

// dump the states for a factor
void FSMDumpStates(FSM_t * this, unsigned int factor) {
  unsigned int i, j;
 
    for (j = 0; j < this->states[factor]; j++) {
      printf("  %s\n", this->stateNames[factor][j]);
    }

}

// find a factor by name
long FSMFindFactor(FSM_t * this, char * name) {
  unsigned int i;
  
  for (i = 0; i < this->factors; i++ ) {
    if (strcmp(name, this->factorNames[i]) == 0)  return i;
  }
  
  return -1;
}

void FSMAddFactorState(FSM_t * this, unsigned int thisFactor, char * name) {

   unsigned int i, j;

   debug("Entered AddFactorState");
   this->states[thisFactor]++;     
   debug(" reallocing memory for new state");  
   this->stateNames[thisFactor]                             = realloc(this->stateNames[thisFactor], (sizeof(char*)*this->states[thisFactor]));
   debug(" copying state name");
   this->stateNames[thisFactor][this->states[thisFactor]-1] = strdup(name);    
   
}

void FSMDumpCode(FSM_t * this) {

   unsigned char bits, totalBits;
   int * maxNameLen;
   int i, j;
      
   FSMinterlinkTable( this );  
      
   maxNameLen = malloc( sizeof(int)*this->factors );

   // this should go, inputs, feedback, ouputs

   printf("\ntypedef struct FSM_State_s { \n");
   totalBits = 0;
   for(i = 0; i < this->factors; i++) {
     if ((this->states[i] > 1)  ){
       for (bits = 0; (1<<bits) < this->states[i]; bits++);
       printf("  unsigned %s : %d;\n", this->factorNames[i], bits);
       totalBits += bits;
     }
   }
   printf("} FSM_State_t;\n");
   
   
   
      printf(" \n\
typedef unsigned int  PTRType; \n\
    \n\
typedef struct stateEntry_s {  \n\
    PTRType       next;            \n\
    FSM_State_t   state;           \n\
 } stateEntry_t;                   \n\
 \n");
 
   // figure out padding for string alignment
    for(i = 0; i < this->factors; i++) {   
      maxNameLen[i] = 0;
      for (j = 0; j < this->states[i]; j++) {   
        if (this->states[i] > 1) {          
          if (strlen(this->stateNames[i][j]) > maxNameLen[i])  maxNameLen[i] = strlen(this->stateNames[i][j]);
        }
      }   
    }
 
   // make C code enumerations
   for(i = 0; i < this->factors; i++) {   
     printf ("//----------| %s |----------\n", this->factorNames[i]);
     for (j = 0; j < this->states[i]; j++) {   
       if (this->states[i] > 1) {          
         printf ("  #define %s ", this->stateNames[i][j]);
         postPad(this->stateNames[i][j], maxNameLen[i]);
         printf("%d\n", j) ;        
       }
     }   
   }
           
    FSMstateTable (this, maxNameLen);
        
}


void FSMinterlinkTable(FSM_t * this) {

   unsigned int i, j;
   
   // calculate number of table lines
   j = 1; for (i = 0; i < this->factors; i++ )  j *= this->states[i]; 
   
   // resize the link table  
   this->linkTable = realloc(this->linkTable, sizeof(unsigned int)*j);
     
   // interlink defults
   for(i = 0; i < j; i++) {
     this->linkTable[i] = i;    // copout
   }
   
   // fancy code here   
   
}


// new output format needs to be
// {feedback_in , inputs}, {putputs, feedback_out}
//
void FSMstateTable (FSM_t * this, int * maxNameLen) {

       int i,j;
       int * stateItter;
       int entries;

       stateItter = malloc( sizeof(int)*this->factors );
       entries = 1;       
       
       for (i = 0; i < this->factors; i++ )  entries *= this->states[i];
       
       printf("//------------\n  // NextState , { ");
       for(i = this->factors - 1; i >= 0; i--) {
         if (this->states[i] > 1) printf ("  %s  ,", this->factorNames[i]);  
       }
       printf("} \n");
              
       printf("stateEntry_t stateTable[] = {   \n");
       
       for (i = 0; i < entries; i++) {
       
         fromIndex(this->states, this->factors, stateItter, i);
                 
         printf("  { %3d, {", this->linkTable[i]); 
         for(j = this->factors - 1; j >= 0; j--) {
          // if ((this->states[j] > 1) & (this->directions[j] != OUTPUT)) {
             printf (" %s ", this->stateNames[j][stateItter[j]]);     
             postPad(this->stateNames[j][stateItter[j]], maxNameLen[j]);
             if (j != 0)printf (",");  
          // }           
         }    
         
         if (i+1 == entries) { printf("}}   /*   index: %u   */",i);   
         } else {              printf("}},  /*   index: %u   */",i); 
         }
         
           printf("\n");
       }
       printf("};\n");
       free(stateItter);
   
}


// forward and reverse table navigation
// take an integer and translate it to details of a table line
void fromIndex(int * Limit, int LimitCount, int * Result, int i) {
  int j, k, n;

  for(j = LimitCount - 1; j >= 0; j--) {
   for (k = 0, n = 1; k < j; k++) n *= Limit[k];
   Result[j] = i / n;
   i -= (Result[j] * n);   
   
  }
}

// take a set of states and tell us what table line it occurs on
int toIndex(int * Limit, int LimitCount, int * CoOrd){
  int i;
  int r;
  int m;
  
  m = 1; 
  r = 0;
   
  for( i = 0; i < LimitCount; i++) {
    r += CoOrd[i] * m;
    m *= Limit[i];
  }   

 return r;
 
}



void test() {
  FSM_t * Project;

  Project = NULL;
  debug("Allocating main project...");  
  Project = realloc(Project, sizeof(FSM_t));
  
  FSMInit(Project);
    
//  printf(" 'foo' is index %d\n", FSMFindFactor(Project, "bar"));

  FSMAddFactor(Project, "bar", INPUT);
  FSMAddFactorState(Project, FSMFindFactor(Project, "bar"), "factor1");
  FSMAddFactorState(Project, FSMFindFactor(Project, "bar"), "fac2");
  
  FSMAddFactor(Project, "that", INPUT);
  FSMAddFactorState(Project, FSMFindFactor(Project, "that"), "facA");
  FSMAddFactorState(Project, FSMFindFactor(Project, "that"), "facB");
  
  FSMAddFactor(Project, "other", OUTPUT);
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2A");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2B");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "factor2C");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2D");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2E");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2F");
  
  FSMDump(Project);
  printf("---------------------------------------------------\n");
  //FSMDelFactorState(Project, FSMFindFactor(Project, "other"), 2);
  FSMDelFactor(Project, FSMFindFactor(Project, "other"));
  
  FSMDump(Project);
  
  FSMAddFactor(Project, "other", INPUT);
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "facA");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "facB");
  
  printf("++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
  
  FSMDump(Project);
  
  //FSMDumpCode(Project);
  
  FSMDestroy(Project); 
}
