/*
             CREATOR: DAn williams
                DATE:
         DESCRIPTION: makes state machine code
            FILENAME: combo3.c
             STARTED:
    OPERATING SYSTEM: Linux
1st VERSION FINISHED:

this code writes state machine code

the states are grouped into banks that are addressed by their inputs, 
so the offset in the state machine is local wihtin the input bank

  factor x
     state 1
     state 2
     state 3
  factor y
     state a
     state b

*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

//#define debug(k)   printf("%s\n", (k));
#define debug(k)  /* (k) */

#define REPLYSZ       512

#define INPUT    0
#define OUTPUT   1
#define FEEDBACK 2

char * DIRECTIONS[] = {"in", "out", "fb" };

#define VERSION 7.0

//global variables

/*
typedef struct FSM_s {  

  char          * filename;   // filename

  unsigned int  factors;      // factors in the state machine (inputs / feedback)
  unsigned int  * states;     // number of states each factor has
  unsigned char * directions; // input or output
  
  char **  factorNames;       // names of factors
  char ***  stateNames;       // names of states
  
  unsigned int * linkTable;   // link to next state (feedback)
  
} FSM_t;           
*/


typedef struct Factor_s {  

  char *         factorName;      // names of factors
  unsigned char  direction;        // input or output
  
  unsigned int   states;           // number of states each factor has
  char **        stateNames;       // names of states

} Factor_t; 


typedef struct Relation_s {

  unsigned int sources;
  unsigned int * srcFactor;
  unsigned int * srcState;
  
  unsigned int dests;
  unsigned int * destFactor;
  unsigned int * destState;

} Relation_t;


typedef struct FSM_s {  

  char        * filename;    // filename

  unsigned int  factorCount; // factors in the state machine (inputs / feedback)  
  Factor_t   ** factors;      // the factors themselvs
  
  unsigned int  linkRuleCount;
  Relation_t ** linkRules;    // state linking rules
  
  unsigned int  outputRuleCount;
  Relation_t ** outputRules;  // outputs for states
  
  unsigned int * linkTable;   // link to next state (feedback)
  
} FSM_t;  


void test();
void makeMachine();
int  doCmd(char * string, FSM_t * Project);
int  getCommand(char ** commands, char ** in) ;

void cmdHelp();
void cmdAddFactor( FSM_t * Project, char * params);
void cmdDelFactor( FSM_t * Project, char * params);
void cmdAddState( FSM_t * Project, char * params);
void cmdDelState( FSM_t * Project, char * params);
void cmdShowStates( FSM_t * Project, char * params);
void cmdFileName( FSM_t * Project, char * params);
void cmdSave( FSM_t * Project, char * params);
void cmdLoad( FSM_t * Project, char * params);

void FSMInit(FSM_t * this) ;
void FSMDestroy(FSM_t * this);
void FSMAddFactor(FSM_t * this, char * name, int direction);
void FSMDelFactorState(FSM_t * this, unsigned int factor, unsigned int state) ;
long FSMFindState(FSM_t * this, unsigned int factor, char * name) ;
void FSMDelFactor(FSM_t * this, unsigned int factor) ; 
void FSMDump(FSM_t * this) ;
void FSMDumpFactors(FSM_t * this);
void FSMDumpStates(FSM_t * this, unsigned int factor);
long FSMFindFactor(FSM_t * this, char * name);
void FSMAddFactorState(FSM_t * this, unsigned int thisFactor, char * name);
void FSMstateTable (FSM_t * this, int * maxNameLen) ;
void FSMDumpCode(FSM_t * this);

void FSMDumpRom(FSM_t * this);  // writeme
void FSMinterlinkTable(FSM_t * this);  //writeme

int  toIndex(int * Limit, int LimitCount, int * CoOrd);
void fromIndex(int * Limit, int LimitCount, int * Result, int i) ;

void printHexString( char * s);
void chomp(char * string);
void postPad(char * string, int finalWidth);







int main(int argc, char** argv) {
  int retCode;
  
  if (isatty(STDIN_FILENO)) {   
    // terminal
  } else {
    // pipe
  }
  
  retCode = 0;
  printf(" This program writes state machines \n");
  printf(" Written by Dan Williams Feb 24 2007 \n");
  printf(" Ver. %f \n", VERSION);
  
  makeMachine();
  //test();
  
  return(retCode);
}

//------------------------------------------------------------------------

/******************************************
   7.0 ok
   Interact with stdin to make a machine
******************************************/

void makeMachine() {

  FSM_t * Project;  
  char * string;
  char reply[REPLYSZ];

  Project = NULL;
  Project = realloc(Project, sizeof(FSM_t));
  FSMInit(Project);
  
  printf(" Type \"help\" for help. \n");
  do {   
    printf("> ");
    fflush(stdout);
    string = fgets(reply, REPLYSZ, stdin);      
  } while((string != NULL) && (!doCmd(string, Project))); 
    
   FSMDestroy(Project); 
}


/*******************************************
   7.0 ok 
   Interpert command string
*******************************************/

int doCmd(char * string, FSM_t * Project) {

// hey doc, these gotta match!
  enum cmdnums { 
         EMPTY = 0,
         HELP,
         ADDFACTOR,
         ADDSTATE,
         BUILD,
         QUIT,
         EXIT,
         SHOW,
         SHOWFACTORS,
         SHOWSTATES,
         FILENAME,
         SAVE,
         LOAD,
         DELFACTOR,
         DELSTATE,
         NEW,
         UNKNOWN
         
  };  
  
  static char * commands[] = { 
         " ",
         "help", 
         "addfactor", 
         "addstate", 
         "build", 
         "quit",
         "exit",
         "show",
         "showfactors",
         "showstates",
         "filename",
         "save",
         "load",
         "delfactor",
         "delstate",
         "new",
         "" 
  };
  
    chomp(string); 
       
    switch( getCommand(commands, &string)) {
      case HELP:
         cmdHelp();
         break;
      case BUILD:
         FSMDumpCode(Project);
         break;
      case ADDFACTOR:  
         cmdAddFactor(Project, string);      
         break;
      case ADDSTATE:
         cmdAddState(Project, string);          
         break;  
      case EXIT:          
      case QUIT:
         return 1;
         break;
      case SHOW:
         FSMDump(Project);  
         break;
      case SHOWFACTORS:
         FSMDumpFactors(Project);    
         break;     
      case SHOWSTATES:
         cmdShowStates(Project, string); 
         break;
      case FILENAME:
         cmdFileName(Project, string);
         break;
      case SAVE:
         cmdSave(Project, string);
         break;
      case LOAD:
         cmdLoad(Project, string);
         break;    
      case NEW:
         FSMDestroy(Project);
         FSMInit(Project);
         break;  
      case DELSTATE:
         cmdDelState(Project, string);
         break;
      case DELFACTOR:
         cmdDelFactor(Project, string);
         break;
      case UNKNOWN:
         printf("unrecognized command.\n");
         break;
    
    }
    
    return 0;

}


/*****************************************
    7.0 ok
    determine which command was entered
*****************************************/

int getCommand(char ** commands, char ** in) {

  char * cursor;
  char * sep = " \t";
  int  i; 

 while( (**in == ' ') || (**in == '\t')) (*in)++;  // skip white space 
 if (**in == '\0') return 0;               // 0 is EMPTY
 cursor = strsep(in, sep); 
 for( i = 0; ((strcmp(cursor, commands[i]) !=0 ) && (*commands[i] != '\0')) ; i++); 
 return i;
 
}


/****************************************
   7.0 ok
   print string in hex digits
****************************************/

void printHexString( char * s) {
  for(;*s ;s++) printf("%02X ", *s);
}

/***************************************
  7.0 ok
  set/get project filename
***************************************/
void cmdFileName( FSM_t * Project, char * params) {
  char * cursor;   
  char * sep = " \t";

  debug("entering cmdfilename");
 
  if ((params==NULL) || (strlen(params) == 0)) {
    if (Project->filename != NULL) 
       printf("%s\n", Project->filename);
    return;
  } else {  
    if (Project->filename != params) { // we tend to get called with our own pointer
      if (Project->filename != NULL) {
        free(Project->filename);      
      } 
      Project->filename = strdup(params);
    }
  } 
  
}

/*************************************
    7.0 ok
    save machine to file
************************************/
void cmdSave( FSM_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   FILE * output;
   unsigned int i, j;
   
  debug("entering cmdsave"); 
  if ((params==NULL) || (strlen(params) == 0) ) {
     if (Project->filename != NULL) {
       debug("calling self with filename");
       cmdSave( Project, Project->filename);
     } else {
       printf ("I need you to give me a filename!\n");
     }
     return;
  } else {   
    debug("seperating commands");
    while( cursor = strsep(&params, sep) ) { 
      if ((output = fopen(cursor, "wb")) == NULL) {
        printf("Couldn't open %s to save machine.\n", cursor);
      } else {
        for (i = 0; i < Project->factorCount; i++ ) {
          debug("adding factor");
          fprintf(output, "addfactor %s %s\n",  DIRECTIONS[Project->factors[i]->direction], Project->factors[i]->factorName);
          for (j = 0; j < Project->factors[i]->states; j++) {
            debug("adding state");
            fprintf(output, "  addstate %s %s\n", Project->factors[i]->factorName, Project->factors[i]->stateNames[j]);
            debug("done adding state");
          }
        }
        debug("closing file");
        fclose(output);
        debug("resetting filename");
        cmdFileName(Project, cursor);
      }                   
    }
  }           
}


/**********************************
  7.0 ok
  load file into current machine
**********************************/

void cmdLoad( FSM_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   FILE * input;
   char * string;
   char reply[REPLYSZ];
   
  if ((params==NULL) || (strlen(params) == 0)) {
    if (Project->filename != NULL) {
       cmdLoad( Project, Project->filename);
     } else {
       printf ("I need you to give me a filename!\n");
     }
     return;
  } else {   
    while( cursor = strsep(&params, sep) ) { 
      if ((input = fopen(cursor, "rb")) == NULL) {
        printf("Couldn't open %s to load machine.\n", cursor);
      } else {      
        do {   
          string = fgets(reply, REPLYSZ, input);      
        } while((string != NULL) && (!doCmd(string, Project)));            
        fclose(input);
        cmdFileName(Project, cursor);
      }                   
    }    
  }           
}


/********************************
   7.0 ok
   show the states of a factor
********************************/

void cmdShowStates( FSM_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   int i;

   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need to know which factor to show you states of!\n");
      return;
   } else {   
      while( cursor = strsep(&params, sep) ) { 
         if ((i = FSMFindFactor(Project, cursor)) == -1) {
           printf ("I cannot find a factor called '%s'.\n", cursor);
         } else {
           printf("%s:\n", cursor);
           FSMDumpStates(Project, i);
         }
      }
   }           
}

/********************************
  7.0 ok
  command to add a state to a factor
********************************/
  
void cmdAddState( FSM_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   int i;

   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need to know which factor to add your state to!\n");
      return;
   } else {
      cursor = strsep(&params, sep);
      if ((i = FSMFindFactor(Project, cursor)) == -1) {
        printf ("I cannot find a factor called '%s'.\n", cursor);
        return;
      }
   }
   
   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need to know what state name(s) you want to add to %s.\n", cursor);
      return;
   }
   
   while( cursor = strsep(&params, sep) ) {      
      FSMAddFactorState(Project, i, cursor); 
   }
}


/*********************************
  7.0 ok
  command to delete a state
*********************************/
  
void cmdDelState( FSM_t * Project, char * params) {
   char * cursor;   
   char * sep = " \t";
   int i, j;

   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need to know which factor to delete your state from!\n");
      return;
   } else {
      cursor = strsep(&params, sep);
      if ((i = FSMFindFactor(Project, cursor)) == -1) {
        printf ("I cannot find a factor called '%s'.\n", cursor);
        return;
      }
   }
   
   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need to know what state names you want to delete from %s.\n", cursor);
      return;
   }
   
   while( cursor = strsep(&params, sep) ) {     
      if ((j = FSMFindState( Project, i, cursor)  ) == -1) {
        printf ("I cannot find a state called %s in factor called '%s'.\n", cursor, Project->factors[i]->factorName);
        return;  
      } else {
        FSMDelFactorState(Project, i, j); 
      }
   }
}


/*****************************
  7.0 ok
  command to add a factor
*****************************/
  
void cmdAddFactor( FSM_t * Project, char * params) {
   char * cursor;
   char * sep = " \t";
   char dir;

   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need a direction and name for the factor you want to add!\n");
      return;
   }
   
   cursor = strsep(&params, sep);
   
   if (strcmp(cursor, "in") == 0) {
     dir = INPUT;
   } else if ( strcmp(cursor, "out") == 0) {
     dir = OUTPUT;
   } else if ( strcmp(cursor, "fb") == 0) {
     dir = FEEDBACK;
   } else {
     printf("direction needs to be 'in', 'fb' or 'out', you said %s.\n", cursor);
     return;
   }
      
   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need a name for the factor you want to add!\n");
      return;
   }   
      
   while( cursor = strsep(&params, sep) ) {
      FSMAddFactor(Project, cursor, dir);
   }
}


/*************************************
  7.0 ok
  comand to delete a factor
*************************************/
  
void cmdDelFactor( FSM_t * Project, char * params) {
   char * cursor;
   char * sep = " \t";
   char dir;
   int i;

   if ((params==NULL) || (strlen(params) == 0)) {
      printf ("I need a  name for the factor you want to delete!\n");
      return;
   }  
   
   while( cursor = strsep(&params, sep) ) {   
      if ((i = FSMFindFactor( Project, cursor) ) == -1) {
        printf ("I cannot find a factor called '%s'.\n", cursor );
        return;  
      } else {
        FSMDelFactor(Project, i); 
      }      
   }
}


/*********************************
  7.0 ok
  print command help
*********************************/

void cmdHelp() {

printf( "\
help                         print help (your looking at it)    \n\
addfactor direction  name(s) add a factor                       \n\
  direction is 'in', 'fb' or 'out'                              \n\
addstate  factor  name       add a state to factor              \n\
showfactors                  list factors                       \n\
showstates factor            list states                        \n\
show                         show all factors and states        \n\
build                         print machine                     \n\
quit                         exit                               \n\
filename {filename}          report or set the current filename \n\
save {filename}                save machine definition          \n\
load {filename}                load state machine definition    \n\
new                          clear project                      \n\
                                                                \n\
delfactor factor             delete factor                      \n\
delstate factor state        delete state from factor           \n\
                                                                \n\
           {} denote optional paramiters.                       \n\
");
}



/***************************************
  7.0 ok
   add padding to output, for alignment
***************************************/

void postPad(char * string, int finalWidth){
 int i;
  for(i = 0; i < finalWidth-strlen(string); i++) { printf(" ");}
}


/**************************************
   7.0 ok
   tear off any trailing linefeeds
**************************************/

void chomp(char * string) {
  char * target;
  
  target = strchr(string,'\n');
  if (target != NULL) { *target = '\0'; }
  
}


//------------------------------------------------- FSM --------------------------------------------------------

/*******************************
  7.0 ok
  initialize a new machine
*******************************/
  
void FSMInit(FSM_t * this) {

  debug("Entering init...");
  this->filename        = NULL;
  
  this->factorCount     = 0;
  this->factors         = NULL;
  
  this->linkRuleCount   = 0;
  this->linkRules       = NULL;
  
  this->outputRuleCount = 0;
  this->outputRules     = NULL;

}

/*******************************
   7.0 ok
   deallocate a machine
*******************************/
void FSMDestroy(FSM_t * this) {
  unsigned int i, j;
 
  debug("Entering fsmdestroy...");
  
  for (i = 0; i < this->factorCount; i++ ) {
    debug("freeing factorname...");
    free(this->factors[i]->factorName); 
    this->factors[i]->factorName = NULL;   
    debug("freeing state names..."); 
    for (j = 0; j < this->factors[i]->states; j++) {
      free(this->factors[i]->stateNames[j]);
    }     
    debug("freeing factor...");
    free(this->factors[i]); 
    this->factors[i] = NULL;
  }
  debug("freeing filename...");
  free(this->filename);    this->filename    = NULL;
  debug("freeing linktable...");
  free(this->linkTable);   this->linkTable   = NULL;
  
  this->factors = 0;
  
}


/***********************************
  7.0 ok
  add a factor to a machine
  Oh really, add some error checking already.
***********************************/

void FSMAddFactor(FSM_t * this, char * name, int direction) {

  this->factorCount++;                                                                               // inc factorcount
  this->factors                                  = realloc(this->factors, (sizeof(Factor_t *)*this->factorCount));    //  allocate factor
  this->factors[this->factorCount-1]             = malloc( sizeof(Factor_t) );
  this->factors[this->factorCount-1]->factorName = strdup(name);
  this->factors[this->factorCount-1]->direction  = direction;
  this->factors[this->factorCount-1]->states     = 0;
  this->factors[this->factorCount-1]->stateNames = NULL;
  
}


/************************************
  v7.0 ok
  delete a state from a factor
***********************************/

void FSMDelFactorState(FSM_t * this, unsigned int factor, unsigned int state) {

 unsigned int i, j;
  debug("entered fsmdelfactorstate");
  if (state <= this->factors[factor]->states) {
  
     debug("freeing state name...");     
     free(this->factors[factor]->stateNames[state]);
     
     debug("moving other state names");     
     for (j = state; j < this->factors[factor]->states - 1; j++) {
          this->factors[factor]->stateNames[j] = this->factors[factor]->stateNames[j+1]; 
     }
          
     this->factors[factor]->states--; 
               
  } 
}


/************************ 
  7.0 ok
  find a state name 
************************/

long FSMFindState(FSM_t * this, unsigned int factor, char * name) {
  unsigned int i, j;
  for (j = 0; j < this->factors[factor]->states; j++) {
    if (strcmp(name, this->factors[factor]->stateNames[j]) == 0)  return j;
  }
  return -1;
}

/***********************************
  7.0 ok
  delete a factor from a machine
***********************************/

void FSMDelFactor(FSM_t * this, unsigned int factor) {
  unsigned int i, j;
  if (factor <= this->factorCount) {
    free(this->factors[factor]->factorName);                    // delete its name
    
    for (j = 0; j < this->factors[factor]->states; j++) {       // delete its factors
      free(this->factors[factor]->stateNames[j]);
    }
  
    free(this->factors[factor]);                                // free the factor itself
  
    for (j = factor; j < this->factorCount - 1; j++) {          // move back the factors      
      this->factors[j] = this->factors[j+1];         
    }        
    
    this->factorCount--;
    this->factors   = realloc(this->factors, (sizeof(Factor_t *)*this->factorCount));    //  allocate factor    
    
  }
}


/*********************************************
   7.0 ok
   dump the states and factors for a machine
*********************************************/

void FSMDump(FSM_t * this) {
  unsigned int i, j;
 
  for (i = 0; i < this->factorCount; i++ ) {
    printf("%s  (%s)\n", this->factors[i]->factorName, DIRECTIONS[this->factors[i]->direction] );
    for (j = 0; j < this->factors[i]->states; j++) {
      printf("  %s\n", this->factors[i]->stateNames[j]);
    }
  }

}


/*************************************
   7.0 ok
   dump the factors for a machine
**************************************/

void FSMDumpFactors(FSM_t * this) {
  unsigned int i, j;
 
  for (i = 0; i < this->factorCount; i++ ) {
    printf("%s\n", this->factors[i]->factorName );
  }

}


/***********************************
  7.0 ok
  dump the states for a factor
***********************************/

void FSMDumpStates(FSM_t * this, unsigned int factor) {
  unsigned int i, j;
 
    for (j = 0; j < this->factors[factor]->states; j++) {
      printf("  %s\n", this->factors[factor]->stateNames[j]);
    }

}


/******************************
   7.0 ok
   find a factor by name
*****************************/

long FSMFindFactor(FSM_t * this, char * name) {
  unsigned int i;
  
  for (i = 0; i < this->factorCount; i++ ) {
    if (strcmp(name, this->factors[i]->factorName) == 0)  return i;
  }
  
  return -1;
}


/***********************
  7.0 ok
  add a state to a factor
***********************/

void FSMAddFactorState(FSM_t * this, unsigned int thisFactor, char * name) {

   unsigned int i, j;

   debug("Entered AddFactorState");
   this->factors[thisFactor]->states++;  
   debug(" reallocing memory for new state");  
   this->factors[thisFactor]->stateNames                    = realloc(this->factors[thisFactor]->stateNames, (sizeof(char*)*this->factors[thisFactor]->states));
   debug(" copying state name");
   this->factors[thisFactor]->stateNames[this->factors[thisFactor]->states - 1] = strdup(name);    
   
}

/****************************
  v7.0 ok
  dump the machine as C code
****************************/
void FSMDumpCode(FSM_t * this) {

   unsigned char bits, totalBits;
   int * maxNameLen;
   int i, j;
      
   FSMinterlinkTable( this );  
      
   maxNameLen = malloc( sizeof(int)*this->factorCount );

   // this should go, ( feedback, inputs ), ( feedback, ouputs )

   printf("\ntypedef struct FSM_Input_s { \n");
   totalBits = 0;
   for(i = 0; i < this->factorCount; i++) {
     if (this->factors[i]->direction != OUTPUT) {
       if ((this->factors[i]->states > 1)  ){
         for (bits = 0; (1<<bits) < this->factors[i]->states; bits++);
         printf("  unsigned %s : %d;\n", this->factors[i]->factorName, bits);
         totalBits += bits;
       }
     }
   }
   printf("} FSM_Input_t;\n");
   
   
   
   printf("\ntypedef struct FSM_Output_s { \n");
   totalBits = 0;
   for(i = 0; i < this->factorCount; i++) {
     if (this->factors[i]->direction != INPUT) {
       if ((this->factors[i]->states > 1)  ){
         for (bits = 0; (1<<bits) < this->factors[i]->states; bits++);
         printf("  unsigned %s : %d;\n", this->factors[i]->factorName, bits);
         totalBits += bits;
       }
     }  
   }
   printf("} FSM_Output_t;\n");
   
   
   
   
      printf(" \n\
typedef struct stateEntry_s {  \n\
    FSM_Input_t       in;            \n\
    FSM_Output_t      out;           \n\
 } stateEntry_t;                   \n\
 \n");
 
   // figure out padding for string alignment
    for(i = 0; i < this->factorCount; i++) {   
      maxNameLen[i] = 0;
      for (j = 0; j < this->factors[i]->states; j++) {   
        if (this->factors[i]->states > 1) {          
          if (strlen(this->factors[i]->stateNames[j]) > maxNameLen[i])  maxNameLen[i] = strlen(this->factors[i]->stateNames[j]);
        }
      }   
    }
 
   // make C code enumerations
   for(i = 0; i < this->factorCount; i++) {   
     printf ("//----------| %s |----------\n", this->factors[i]->factorName);
     for (j = 0; j < this->factors[i]->states; j++) {   
       if (this->factors[i]->states > 1) {          
         printf ("  #define %s ", this->factors[i]->stateNames[j]);
         postPad(this->factors[i]->stateNames[j], maxNameLen[i]);
         printf("%d\n", j) ;        
       }
     }   
   }
           
    FSMstateTable (this, maxNameLen);
        
}


/**************************************************
  v7.0 ok
  new output format needs to be
  {feedback_in , inputs}, {putputs, feedback_out}
****************************************************/

void FSMstateTable (FSM_t * this, int * maxNameLen) {

       int i,j;
       int * stateItter;
       int entries;
       int * stateCounts;
       int inputFactorCount;
       int * factorList;

       stateItter  = malloc( sizeof(int)*this->factorCount );
       stateCounts = malloc( sizeof(int)*this->factorCount );
       factorList  = malloc( sizeof(int)*this->factorCount );
       entries     = 1;       
       inputFactorCount = 0;
       
       for (i = 0; i < this->factorCount; i++ ) {
        if (this->factors[i]->direction != OUTPUT) {         
         entries *= this->factors[i]->states;
         stateCounts[inputFactorCount] = this->factors[i]->states;         
         factorList[inputFactorCount] = i;   
         inputFactorCount++; 
        }
       }
       
       // table header
       printf("//------------\n  // NextState , { ");
       for(i = this->factorCount - 1; i >= 0; i--) {
         if (this->factors[i]->states > 1) printf ("  %s  ,", this->factors[i]->factorName);  
       }
       printf("} \n");
              
       // main state table       
       printf("stateEntry_t stateTable[] = {   \n");
       
       for (i = 0; i < entries; i++) {
       
         fromIndex(stateCounts, inputFactorCount, stateItter, i);
                 
       //  printf("  { %3d, {", this->linkTable[i]); 
         for(j = inputFactorCount - 1; j >= 0; j--) {
             printf (" %s ", this->factors[factorList[j]]->stateNames[stateItter[j]]);     
             postPad(this->factors[factorList[j]]->stateNames[stateItter[j]], maxNameLen[factorList[j]]);
             if (j != 0)printf (",");          
         }    
         
         if (i+1 == entries) { printf("}}   /*   index: %u   */",i);   
         } else {              printf("}},  /*   index: %u   */",i); 
         }
         
           printf("\n");
       }
       printf("};\n");
       free(stateItter);
       free(stateCounts);
   
}

/***************************** 
  7.0 ok
  calculate the feedback links for the table
******************************/

void FSMinterlinkTable(FSM_t * this) {

   unsigned int i, j;
   
   // calculate number of table lines
   j = 1; for (i = 0; i < this->factorCount; i++ )  j *= this->factors[i]->states; 
   
   // resize the link table  
   this->linkTable = realloc(this->linkTable, sizeof(unsigned int)*j);
     
   // interlink defults
   for(i = 0; i < j; i++) {
     this->linkTable[i] = i;    // copout
   }
   
   // fancy code here   
   
}


/************************************
 7.0 ok
 forward and reverse table navigation
 take an integer and translate it to details of a table line
************************************/

void fromIndex(int * Limit, int LimitCount, int * Result, int i) {
  int j, k, n;

  for(j = LimitCount - 1; j >= 0; j--) {
   for (k = 0, n = 1; k < j; k++) n *= Limit[k];
   Result[j] = i / n;
   i -= (Result[j] * n);   
   
  }
}

/*************************************
  0.7 ok
  take a set of states and tell us what table line it occurs on
************************************/

int toIndex(int * Limit, int LimitCount, int * CoOrd){
  int i, r, m;

  m = 1;  r = 0;
   
  for( i = 0; i < LimitCount; i++) {
    r += CoOrd[i] * m;
    m *= Limit[i];
  }   

 return r;
 
}


/******************************
  7.0 ok
  basic test of functionality
******************************/

void test() {
  FSM_t * Project;

  Project = NULL;
  debug("Allocating main project...");  
  Project = realloc(Project, sizeof(FSM_t));
  
  FSMInit(Project);
    
//  printf(" 'foo' is index %d\n", FSMFindFactor(Project, "bar"));

  FSMAddFactor(Project, "bar", INPUT);
  FSMAddFactorState(Project, FSMFindFactor(Project, "bar"), "factor1");
  FSMAddFactorState(Project, FSMFindFactor(Project, "bar"), "fac2");
  
  FSMAddFactor(Project, "that", INPUT);
  FSMAddFactorState(Project, FSMFindFactor(Project, "that"), "facA");
  FSMAddFactorState(Project, FSMFindFactor(Project, "that"), "facB");
  
  FSMAddFactor(Project, "other", OUTPUT);
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2A");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2B");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "factor2C");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2D");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2E");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "fac2F");
  
  FSMDump(Project);
  printf("---------------------------------------------------\n");
  //FSMDelFactorState(Project, FSMFindFactor(Project, "other"), 2);
  FSMDelFactor(Project, FSMFindFactor(Project, "other"));
  
  FSMDump(Project);
  
  FSMAddFactor(Project, "other", INPUT);
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "facA");
  FSMAddFactorState(Project, FSMFindFactor(Project, "other"), "facB");
  
  printf("++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
  
  FSMDump(Project);
  
  //FSMDumpCode(Project);
  
  FSMDestroy(Project); 
}
